//************************************************************
//  University of Patras
//	Department of Electrical Engineering & Computer Technology
//
//	Course: Parallel Programming in AI
//	Week 05: OpenMp-OpenACC
//
//  Description: Sort using openACC. Parallel Pair Sort in GPU
//	Author: Evangelos Dermatas
//	v0.1
/*
*********************************************************************
gcc -v:	gcc version 7.4.0 (Ubuntu 7.4.0-1ubuntu1~18.04.1)
OS: 	Linux Ubuntu 18.04
CPU:	8-core, Intel(R) Core(TM) i7-2630QM CPU @ 2.00GHz, 3991 BogoMips
*********************************************************************
Compiling:	gcc accSort06.c -o ompSort06 -fopenacc
Executing:	./ompSort06

FOR N=100000
0 0 16331 21720 32534 43513 48959 65535

real	0m3.207s
user	0m25.465s
sys	0m0.088s

*/
//************************************************************


// *******************************************************************
//#pragma GCC optimize("O3")
//#pragma GCC option("arch=native","openacc","offload=nvptx-none","offload=O3","opt-info")
//************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openacc.h>

// Array Size
#define N          100000

unsigned short int A[N] ;

// ***************************************
unsigned long createArray( void )
{
for (unsigned int i=0 ; i < N ; i++ )
	A[i] = rand() % 65536 ;
}


// ***************************************
void PrintResults()
{
for( unsigned int i = 0 ; i < N ; i++ )
	printf( "%hu ", A[i]) ;
putchar('\n') ;
}

// ***************************************
void PairSort()
{
int c ;
int rep = 0 ;
#pragma acc data copy(A[0:N])
{

do {
	c = 0 ;
#pragma acc parallel loop independent present(A)
	for( unsigned int j = 0 ; j < N ; j+=2 )
		if ( A[j] > A[j+1] ) {
		unsigned short k = A[j] ;
		A[j]=A[j+1] ;
		A[j+1]=k ;
		}

#pragma acc parallel loop independent present(A)
	for( unsigned int j = 1 ; j < N-1 ; j+=2 )
		if ( A[j] > A[j+1] ) {
		unsigned short k = A[j] ;
		A[j]=A[j+1] ;
		A[j+1]=k ;
		}

if ( !(rep++ & 0xF) ) {
	int k = 0 ;
#pragma acc parallel loop independent present(A) reduction(max:k) copy(k)
	for( unsigned int j = 1 ; j < N-1 ; j++ )
		if ( A[j] > A[j+1] )
			k = 1 ;
	c = k ;
	}
else
	c = 1 ;

} while(c) ;
}
}

// ***************************************
int checkArray()
{
for( int i = 0 ; i < N-1 ; i++ )
	if ( A[i] > A[i+1] )
		return 1 ;
return 0 ;
}

// ***************************************
int main()
{

createArray() ;

PairSort() ;

printf( "%hu %hu %hu %hu ",  A[0], A[1], A[N/4], A[N/3] ) ;
printf( "%hu %hu %hu %hu\n", A[N/2], A[2*N/3], A[3*N/4], A[N-1] ) ;
//if ( checkArray() )
//	printf( "Error in sorting\n" ) ;
return 0 ;
}
